static int debug = 0;

/*****************************************************************
 * Copyright (c) ParaSoft Corporation 1991.  All rights reserved.*
 * THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF ParaSoft       *
 * The copyright notice above does not evidence any              *
 * actual or intended publication of such source code.           *
 *****************************************************************/

#include "extern.h"

int source_form = FORM_FIXED;

char *form_old_line = (char*) 0;
int form_old_len = -1;  /* means unused */

static int form_stmt_lines = 0;

form_set (form)
char *form;
{
	if (!strcmp (form, "fixed") || !strcmp (form, "FIXED")) {
		source_form = FORM_FIXED;
	}
	else if (!strcmp (form, "free") || !strcmp (form, "FREE")) {
		source_form = FORM_FREE;
	}
	else {
		sprintf (serr, "f90: invalid source form: %s", form);
		terminatenl (serr);
	}
}

form_input (buf, result, max_size, yyin)
char *buf;
int *result, max_size;
FILE *yyin;
{
	int i;

next_line:

	U_stack_clear_key_list ();
	U_stack_clear_strings_list ();
	key_current = 0;
	strings_current = 0;
	form_stmt_lines = 0;

	switch (source_form) {
	case FORM_FIXED:
		form_input_fixed (buf, result, max_size, yyin);
		break;
	case FORM_FREE:
		form_input_free (buf, result, max_size, yyin);
		break;
#ifdef CHECK
	default:
		SYSERR;
#endif
	}

	if (*result == 0)
		return;

	if (debug)
		printf ("1 buf = [%s]\n", buf);

	if (strings_find (buf))
		goto next_line;

	if (debug)
		printf ("2 buf = [%s]\n", buf);

	key_find_keywords (buf);

	if (debug)
		printf ("3 buf = [%s]\n", buf);

	if (*result)
		*result = strlen (buf);
}

#define LFIXED_INITIAL		1
#define LFIXED_CONTINUATION	2
#define LFIXED_ERROR		3
#define LFIXED_COMMENT		4

static int check_fixed_format (line, len, report_error)
char *line;
int len;
int report_error;
{
	static char *nostand = "non-standard source format";
	static char *badform = "bad formatted line";
	int i;
	
	if (line == (char*) 0)  /* EOF */
		return LFIXED_INITIAL;

	if (len == 0)  /* it is just an empty line */
		return LFIXED_COMMENT;

	if (line [0] == 'C' || line [0] == 'c' || line [0] == '*' || line [0] == '!')
/* ##BEG##, 9/19/91, SHU */
		/* save the compiler directives */
		{
		if (line[1] == '$' || line[1] == 'Z')  /* ZEKI */
		    { line[0] = ' '; line[1] = ' '; return LFIXED_INITIAL; }
		else
/* ##END## */
		return LFIXED_COMMENT;
		}

	if (line [0] == 'D' || line [0] == 'd' || line [0] == '$' || line [0] == '') {
		if (report_error) {
			sprintf (serr, "line starts with '%c', assuming comment line", line [0]);
			warning (serr);
		}

		return LFIXED_COMMENT;
	}

	if (!isdigit (line [0]) && !iswhite (line [0])) {
		if (report_error)
			fatal (badform);
		return LFIXED_ERROR;
	}

	if (line [0] == '\t' || line [1] == '\t' || line [2] == '\t'
	 || line [3] == '\t' || line [4] == '\t' || line [5] == '\t') {
		stand (nostand);
		return LFIXED_INITIAL;
	}

	if (len > 5 && line [5] != ' ' && line [5] != '0') {  /* tries to be a cont. line */
		for (i = 0; i < 5; i++)
			if (line [i] != ' ') {  /* checks if 1-5 are all blanks */
				if (report_error)
					fatal (badform);
				return LFIXED_ERROR;
			}
		return LFIXED_CONTINUATION;
	}

/* if it got here is trying to be an standard initial line */

	for (i = 0; i < U_MIN (5, len); i++) 
		if (line [i] != ' ' && !isdigit (line [i])) {
			if (report_error)
				fatal (badform);
			return LFIXED_ERROR;
		}

/* and column 6 must have ' ' or '0' */

	if (len > 5 && line [5] != ' ' && line [5] != '0') {
		if (report_error)
			fatal (badform);
		return LFIXED_ERROR;
	}

	if (len > 5)
		line [5] = ' ';

	for (i = 0; i < len; i++)
		if (!iswhite (line [i]))
			return LFIXED_INITIAL;

/* if it gets here all where whites */

	return LFIXED_COMMENT;
}

/*!!! Things left:
1) the END stuff. (3-6.31)
*/

static form_input_fixed (buf, result, max_size, yyin)
char *buf;
int *result, max_size;
FILE *yyin;
{
	char *stmt = (char*) 0;
	static int stmt_len = 0;
	char *line, num_lines [10];
	int len, check, num_cont_lines;
	int line_len;

	if (flag_e)
		line_len = 132;
	else
		line_len = 72;

next_line:

	line = getline (yyin, &len, line_len);

	form_stmt_lines++;

	if (line == (char*) 0) {  /* got to EOF */
		*result = 0;
		return;
	}

	check = check_fixed_format (line, len, 1);

	switch (check) {

	case LFIXED_CONTINUATION:
		fatal ("illegal continuation card");
		line [5] = ' ';
		break;

	case LFIXED_INITIAL:
		break;

	case LFIXED_COMMENT:
	case LFIXED_ERROR:
		zzlineno++;
		form_stmt_lines--;
		U_free (line);
		goto next_line;

#ifdef CHECK
	default:
		SYSERR;
#endif
	}

	stmt = U_strcat (stmt, line, &stmt_len, 200);
	U_free (line);

	num_cont_lines = 0;

#ifdef CHECK
	if (form_stmt_lines != 1)
		SYSERR;
#endif

continuation:  /* adds continuation lines */

	line = getline (yyin, &len, line_len);

	check = check_fixed_format (line, len, 0);

	switch (check) {

	case LFIXED_CONTINUATION:

		form_stmt_lines++;

		if (++num_cont_lines == 20)
			stand ("fixed form statement has more than 19 continuation lines");
		 
		stmt = U_strcat (stmt, &line [6], &stmt_len, 200);
		U_free (line);
		goto continuation;

	case LFIXED_INITIAL:

terminate_line:

		ungetline (line, len, yyin);

		stmt = U_strcat (stmt, "\n", &stmt_len, 10);
		sprintf (num_lines, "%d\n", form_stmt_lines);
		form_stmt_lines = 0;
		stmt = U_strcat (stmt, num_lines, &stmt_len, 10);
		strcpy (buf, stmt);

		*result = strlen (stmt);

#ifdef CHECK
		if (*result > max_size) {
			SYSERR;
		}
#endif

		U_free (stmt);
		stmt = (char*) 0;
		stmt_len = 0;

		break;

	case LFIXED_COMMENT:

		form_stmt_lines++;
		U_free (line);
		goto continuation;
		break;

	case LFIXED_ERROR:

		goto terminate_line;

#ifdef CHECK
	default:
		SYSERR;
#endif
	}
}

#define LFREE_SIMPLE	1
#define LFREE_BEG		2
#define LFREE_END		3
#define LFREE_BEG_END	4
#define LFREE_SINGLE	5
#define LFREE_END_CHAR	6  /*!!! not implemented yet */
#define LFREE_COMMENT	7

static int check_free_format (pline, plen, report_error, pbeg)
char **pline;
int *plen;
int report_error;
int *pbeg;
{
	int i, beg = -1, end = -1;
	char *line = *pline, *newline;
	int len = *plen;

	*pbeg = 0;

	for (i = 0; i < len; i++)  /* looks at first non-blank character */
		switch (line [i]) {
		case ' ':
		case '\t':
			break;

		case '&':
			beg = i;
			goto lab1;

		default:
			goto lab1;
		}

	return LFREE_COMMENT;  /* if all blanks is comment, !comment removed by getline */

lab1:

	for (i = len - 1; i >= 0; i--)  /* looks at last non-blank character */
		switch (line [i]) {
		case ' ':
		case '\t':
			break;

		case '&':
			end = i;
			goto lab2;

		default:
			goto lab2;
		}

	SYSERR;

lab2:

	*pbeg = beg;

	if (end >= 0) {
		newline = (char*) U_malloc (end + 1);
		line [end] = '\0';
		strcpy (newline, line);
		U_free (line);
		*plen = end;
		*pline = newline;
	}

	if (beg < 0 && end < 0)
		return LFREE_SIMPLE;

	if (beg == end) {
		stand ("single '&' in a line is not allowed");
		return LFREE_SINGLE;
	}

	if (beg >= 0 && end < 0)
		return LFREE_BEG;

	if (beg < 0 && end >= 0)
		return LFREE_END;

	if (beg >= 0 && end >= 0)
		return LFREE_BEG_END;
}

/*!!! Things left:
1) If token split, or character context continuation ==> needs pair of &s
*/

static form_input_free (buf, result, max_size, yyin)
char *buf;
int *result, max_size;
FILE *yyin;
{
	char *stmt = (char*) 0;
	static int stmt_len = 0;
	char *line, num_lines [10];
	int len, check, prev_check, num_cont_lines;
	int beg;

	form_stmt_lines = 0;

next_line:

	line = getline (yyin, &len, 132);
	form_stmt_lines++;

	if (line == (char*) 0) {  /* got to EOF */
		*result = 0;
		return;
	}

	check = check_free_format (&line, &len, 1, &beg);

	switch (check) {

	case LFREE_SIMPLE:
	case LFREE_END:
	case LFREE_END_CHAR:
		break;

	case LFREE_BEG:
	case LFREE_BEG_END:
	case LFREE_SINGLE:
		fatal ("illegal continuation card");
		/* no break... */

	case LFREE_COMMENT:
		zzlineno++;
		form_stmt_lines--;
		U_free (line);
		goto next_line;

#ifdef CHECK
	default:
		SYSERR;
#endif
	}

	stmt = U_strcat (stmt, line, &stmt_len, 200);
	U_free (line);

	if (check == LFREE_SIMPLE)
		goto terminate_line_no_unget;

	num_cont_lines = 0;

#ifdef CHECK
	if (form_stmt_lines != 1)
		SYSERR;
#endif

continuation:  /* adds continuation lines */

	line = getline (yyin, &len, 132);

	prev_check = check;
	check = check_free_format (&line, &len, 0, &beg);

	switch (check) {

	case LFREE_END:
	case LFREE_END_CHAR:
	case LFREE_SIMPLE:

		if (prev_check == LFREE_END_CHAR)
			stand ("character context requires continuation line to start with '&'");

	case LFREE_BEG_END:
	case LFREE_BEG:

		form_stmt_lines++;

		if (++num_cont_lines == 40)
			stand ("free form statement has more than 39 continuation lines");
		 
		stmt = U_strcat (stmt, &line [beg + 1], &stmt_len, 200);
		U_free (line);

		if (check == LFREE_BEG || check == LFREE_SIMPLE)
			goto terminate_line_no_unget;
		else
			goto continuation;

	case LFREE_COMMENT:
	case LFREE_SINGLE:

		form_stmt_lines++;

		U_free (line);
		goto continuation;
		break;

#ifdef CHECK
	default:
		SYSERR;
#endif
	}

#ifdef CHECK
	SYSERR;
#endif

terminate_line:

		ungetline (line, len, yyin);

terminate_line_no_unget:

		stmt = U_strcat (stmt, "\n", &stmt_len, 10);
		sprintf (num_lines, "%d\n", form_stmt_lines);
		stmt = U_strcat (stmt, num_lines, &stmt_len, 10);
		strcpy (buf, stmt);
		*result = strlen (stmt);

#ifdef CHECK
		if (*result > max_size) {
			SYSERR;
		}
#endif

		U_free (stmt);
		stmt = (char*) 0;
		stmt_len = 0;
}

/* gets a line from yyin, (char*) 0 is returned upon end of file, len is the actual
   length of the line */

ungetline (line, len, yyin)
char *line;
int len;
FILE *yyin;
{
#ifdef CHECK
	if (form_old_len >= 0) {
		sprintf (serr, "line [%s]", form_old_line);
		syserr (serr);
	}
#endif

	form_old_line = line;
	form_old_len = len;
}

char *getline (yyin, len, max_len)
FILE *yyin;
int *len, max_len;
{
	int i, state = 0;
	register int c;
	char *cs, *line;

	int ih, j;
	char chlen [9];
	static int hlen = 0;

	if (form_old_len >= 0) { 
		line = form_old_line;
		*len = form_old_len;
		form_old_len = -1;

		return line;
	}

next_line:

	cs = line = (char*) U_malloc (max_len + 1);
	*len = 0;

	for (i = 0; i < max_len; i++) {
		c = fgetc (yyin);

		switch (state) {

		case 0:
			if (!(is_xalpha (c) || iswhite (c)))
				state = -1;
			break;

		case -1:
			if (isdigit (c)) {
				ih = i;
				state = -2;
			}
			break;

		case -2:
			if (c == 'h' || c == 'H') {
				if (is_just_decl (line, line + ih)
				||  is_just_label (line, line + ih)) {
					state = 0;
					break;
				}

				for (j = 0; ih < i; ih++)
					if (!iswhite (line [ih]))
						chlen [j++] = line [ih];

				chlen [j] = '\0';

				hlen = atoi (chlen);
				state = -3;
				goto add_char;
			}
			break;
		}

		if (state < 0 && state > -3) {
			if (is_xalpha (c))
				state = 0;
			else if (!(is_xalpha (c) || iswhite (c) || isdigit (c)))
				state = -1;
		}

		switch (c) {

		case EOF:

			if (i) {
				*len = 0;
				goto found_newline;
			}
			else {
				*len = 0;
				return (char*) 0;
			}
			break;

		case '\n':
		case '\0':

			goto found_newline;
			break;

		case '\'':

			switch (state) {
			case 0:
			case -1:
			case -2:
				state = 1;
				break;
			case 1:
				state = 0;
				break;
			}
			goto add_char;

		case '"':

			switch (state) {
			case 0:
			case -1:
			case -2:
				state = 2;
				break;
			case 2:
				state = 0;
				break;
			}
			goto add_char;

		case '!':

			if (i == 5 && source_form == FORM_FIXED
			&& line [0] == ' '
			&& line [1] == ' '
			&& line [2] == ' '
			&& line [3] == ' '
			&& line [4] == ' ')
				goto add_char;
			if (state < 0 && state > -3)  /* is comment */
				goto eat_rest;
			goto add_char;

		default:
		add_char:
			*cs++ = c;
			(*len)++;
		}

		if (hlen-- == 0)
			state = 0;
	}

eat_rest:

	while (fgetc (yyin) != '\n')
		;

found_newline:

	*cs = '\0';

	return line;
}
